@@ -1,8 +1,9 @@
+
 /*
  * $Id: irc.c,v 1.156 2005/04/21 06:58:50 nohar Exp $
  *
  * This file is part of the bip project
- * Copyright (C) 2004 2005 Arnaud Cornet and Loc Gomez
+ * Copyright (C) 2004 2005 Arnaud Cornet and Loïc Gomez
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,10 +22,12 @@
 #include "log.h"
 #include "connection.h"
 #include "md5.h"
+#include "version.h"
 
 #define S_CONN_DELAY (10)
 
 extern int sighup;
+extern bip_t *_bip;
 
 static int irc_join(struct link_server *server, struct line *line);
 static int irc_part(struct link_server *server, struct line *line);
@@ -45,22 +48,26 @@ static int irc_368(struct link_server *server, struct line *l);
 void irc_server_shutdown(struct link_server *s);
 static int origin_is_me(struct line *l, struct link_server *server);
 
+#ifdef HAVE_OIDENTD
+#define OIDENTD_FILENAME ".oidentd.conf"
+void oidentd_dump(bip_t *bip);
+#endif
+
 void irc_client_free(struct link_client *cli);
-extern int conf_backlog;
 extern int conf_log_sync_interval;
-extern int conf_error;
-extern char conf_errstr[];
-extern int conf_blreset_on_talk;
 
 void write_user_list(connection_t *c, char *dest);
 
 static void irc_copy_cli(struct link_client *src, struct link_client *dest,
 		struct line *line);
 static void irc_cli_make_join(struct link_client *ic);
+static void server_setup_reconnect_timer(struct link *link);
+int irc_cli_bip(bip_t *bip, struct link_client *ic, struct line *line);
 
-#define LAGOUT_TIME (360)
+#define LAGOUT_TIME 480
 #define LAGCHECK_TIME (90)
 #define RECONN_TIMER (120)
+#define RECONN_TIMER_MAX (600)
 #define LOGGING_TIMEOUT (360)
 #define CONN_INTERVAL 60
 #define CONNECT_TIMEOUT 60
@@ -216,7 +223,7 @@ static void irc_server_connected(struct link_server *server)
         LINK(server)->s_state = IRCS_CONNECTED;
         LINK(server)->s_conn_attempt = 0;
 
-	mylog(LOG_INFO, "Connected user %s to %s", LINK(server)->username,
+	mylog(LOG_INFO, "Connected user %s to %s", LINK(server)->user->name,
 		LINK(server)->name);
 
         irc_server_join(server);
@@ -338,9 +345,12 @@ void rotate_who_client(struct link *link)
  * parses:join part mode kick kill privmsg quit nick names
  * returns: -1 invalid protocol
  */
-int irc_dispatch_server(struct link_server *server, struct line *line)
+int irc_dispatch_server(bip_t *bip, struct link_server *server,
+		struct line *line)
 {
 	int ret = OK_COPY;
+	/* shut gcc up */
+	(void)bip;
 
 	if (line->elemc == 0)
 		return ERR_PROTOCOL;
@@ -373,11 +383,16 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 			size_t nicklen = strlen(server->nick);
 			char *newnick = malloc(nicklen + 2);
 			strcpy(newnick, server->nick);
-			if (strlen(server->nick) < 9)
+			if (strlen(server->nick) < 9) {
 				strcat(newnick, "`");
-			else if (newnick[8] != '`') {
-				newnick[8] = '`';
-				newnick[9] = 0;
+			} else if (newnick[7] != '`') {
+				if (newnick[8] != '`') {
+					newnick[8] = '`';
+					newnick[9] = 0;
+				} else {
+					newnick[7] = '`';
+					newnick[9] = 0;
+				}
 			} else {
 				newnick[8] = rand() * ('z' - 'a') / RAND_MAX +
 					'a';
@@ -485,6 +500,11 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 static void irc_send_join(struct link_client *ic, struct channel *chan)
 {
 	char *ircmask; /* fake an irc mask for rbot */
+	struct user *user;
+
+	user = LINK(ic)->user;
+	if (!user)
+		fatal("irc_send_join: No user associated");
 
 	ircmask = malloc(strlen(LINK(ic)->l_server->nick) +
 			strlen("!bip@bip.bip.bip") + 1);
@@ -500,7 +520,10 @@ static void irc_send_join(struct link_client *ic, struct channel *chan)
 				chan->name, chan->creator, chan->create_ts);
 
 	/* XXX: could be more efficient */
-	if (conf_backlog && log_has_backlog(LINK(ic)->log, chan->name)) {
+	if (!user->backlog) {
+		mylog(LOG_DEBUG, "Backlog disabled for %s, not backlogging",
+			user->name);
+	} else if (log_has_backlog(LINK(ic)->log, chan->name)) {
 		char *line;
 		int skip = 0;
 		while ((line =
@@ -511,6 +534,9 @@ static void irc_send_join(struct link_client *ic, struct channel *chan)
 		}
 		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", chan->name,
 				"End of backlog.");
+	} else {
+		mylog(LOG_DEBUG, "Nothing to backlog for %s/%s",
+			user->name, chan->name);
 	}
 
 	list_t *name_list = channel_name_list(chan);
@@ -527,7 +553,6 @@ static void irc_send_join(struct link_client *ic, struct channel *chan)
 
 	WRITE_LINE3(CONN(ic), P_SERV, "366", LINK(ic)->l_server->nick,
 			chan->name, "End of /NAMES list.");
-
 }
 
 static void write_init_string(connection_t *c, struct line *line, char *nick)
@@ -568,14 +593,9 @@ void unbind_from_link(struct link_client *ic)
 		fatal("unbind_from_link");
 
 	if (l->who_client == ic) {
-		mylog(LOG_DEBUG, "unbind_from_link:  %p: %d",
-				l->who_client, ic->who_count);
-		l->who_client = NULL;
-	} else {
-		mylog(LOG_DEBUG,
-			"unbind_from_link: nothing to do %p != %p: %d",
-				ic, l->who_client,
+		mylog(LOG_DEBUG, "unbind_from_link:  %p: %d", l->who_client,
 				ic->who_count);
+		l->who_client = NULL;
 	}
 
 	for (i = i + 1; i < l->l_clientc; i++)
@@ -592,9 +612,9 @@ void unbind_from_link(struct link_client *ic)
 		fatal("realloc");
 }
 
-int irc_cli_bip(struct link_client *ic, struct line *line)
+int irc_cli_bip(bip_t *bip, struct link_client *ic, struct line *line)
 {
-	return adm_bip(ic, line, 0);
+	return adm_bip(bip, ic, line, 0);
 }
 
 #define PASS_SEP ':'
@@ -667,8 +687,8 @@ static void irc_cli_make_join(struct link_client *ic)
 	}
 }
 
-static int irc_cli_startup(struct link_client *ic, struct line *line,
-		list_t *linkl)
+static int irc_cli_startup(bip_t *bip, struct link_client *ic,
+		struct line *line)
 {
 	char *init_nick;
 	char *user, *pass, *connname;
@@ -693,11 +713,13 @@ static int irc_cli_startup(struct link_client *ic, struct line *line,
 	}
 
 	list_iterator_t it;
-	for (list_it_init(linkl, &it); list_it_item(&it); list_it_next(&it)) {
+	for (list_it_init(&bip->link_list, &it); list_it_item(&it);
+			list_it_next(&it)) {
 		struct link *l = list_it_item(&it);
-		if (strcmp(user, l->username) == 0
-				&& strcmp(connname, l->name) == 0) {
-			if (chash_cmp(pass, l->password, l->seed) == 0) {
+		if (strcmp(user, l->user->name) == 0 &&
+				strcmp(connname, l->name) == 0) {
+			if (chash_cmp(pass, l->user->password,
+						l->user->seed) == 0) {
 				bind_to_link(l, ic);
 				break;
 			}
@@ -743,6 +765,7 @@ static int irc_cli_startup(struct link_client *ic, struct line *line,
 		return OK_CLOSE;
 	}
 
+	list_remove(&bip->connecting_client_list, ic);
 	TYPE(ic) = IRC_TYPE_CLIENT;
 
 	for (list_it_init(&LINK(ic)->init_strings, &it);
@@ -779,7 +802,7 @@ static int irc_cli_startup(struct link_client *ic, struct line *line,
 	return OK_FORGET;
 }
 
-static int irc_cli_nick(struct link_client *ic, struct line *line, list_t *cl)
+static int irc_cli_nick(bip_t *bip, struct link_client *ic, struct line *line)
 {
 	if (line->elemc != 2)
 		return ERR_PROTOCOL;
@@ -793,16 +816,17 @@ static int irc_cli_nick(struct link_client *ic, struct line *line, list_t *cl)
 	ic->init_nick = strdup(line->elemv[1]);
 
 	if ((ic->state & IRCC_READY) == IRCC_READY)
-		return irc_cli_startup(ic, line, cl);
+		return irc_cli_startup(bip, ic, line);
 
 	if ((ic->state & IRCC_PASS) != IRCC_PASS)
 		WRITE_LINE2(CONN(ic), P_SERV, "NOTICE", ic->init_nick,
-				"You should type /QUOTE PASS your_username:your_password:your_connection_name");
+				"You should type /QUOTE PASS your_username:"
+				"your_password:your_connection_name");
 
 	return OK_FORGET;
 }
 
-static int irc_cli_user(struct link_client *ic, struct line *line, list_t *cl)
+static int irc_cli_user(bip_t *bip, struct link_client *ic, struct line *line)
 {
 	if (line->elemc != 5)
 		return ERR_PROTOCOL;
@@ -812,11 +836,11 @@ static int irc_cli_user(struct link_client *ic, struct line *line, list_t *cl)
 
 	ic->state |= IRCC_USER;
 	if ((ic->state & IRCC_READY) == IRCC_READY)
-		return irc_cli_startup(ic, line, cl);
+		return irc_cli_startup(bip, ic, line);
 	return OK_FORGET;
 }
 
-static int irc_cli_pass(struct link_client *ic, struct line *line, list_t *cl)
+static int irc_cli_pass(bip_t *bip, struct link_client *ic, struct line *line)
 {
 	if (line->elemc != 2)
 		return ERR_PROTOCOL;
@@ -829,7 +853,7 @@ static int irc_cli_pass(struct link_client *ic, struct line *line, list_t *cl)
 		free(ic->init_pass);
 	ic->init_pass = strdup(line->elemv[1]);
 	if ((ic->state & IRCC_READY) == IRCC_READY)
-		return irc_cli_startup(ic, line, cl);
+		return irc_cli_startup(bip, ic, line);
 	return OK_FORGET;
 }
 
@@ -840,16 +864,17 @@ static int irc_cli_quit(struct link_client *ic, struct line *line)
 	return OK_CLOSE;
 }
 
-static int irc_cli_privmsg(struct link_client *ic, struct line *line)
+static int irc_cli_privmsg(bip_t *bip, struct link_client *ic,
+		struct line *line)
 {
 	if (line->elemc >= 3)
 		log_cli_privmsg(LINK(ic)->log, LINK(ic)->l_server->nick,
 				line->elemv[1], line->elemv[2]);
 	if (strcmp(line->elemv[1], "-bip") == 0)
-		return adm_bip(ic, line, 1);
+		return adm_bip(bip, ic, line, 1);
 
-	if (conf_blreset_on_talk)
-		adm_blreset(ic);
+	if (LINK(ic)->user->blreset_on_talk)
+		log_reinit_all(LINK(ic)->log);
 	return OK_COPY_CLI;
 }
 
@@ -926,8 +951,7 @@ static void irc_notify_disconnection(struct link_server *is)
 					c->name, is->nick,
 					"Server disconnected, reconnecting");
 		}
-		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", is->nick,
-					"Server disconnected, reconnecting");
+		bip_notify(ic, "Server disconnected, reconnecting");
 	}
 }
 
@@ -940,9 +964,10 @@ void irc_add_channel_info(struct link_server *ircs, char *chan, char *key)
 	ci = hash_get(&LINK(ircs)->chan_infos, chan);
 	if (!ci) {
 		struct chan_info *ci;
-		ci = malloc(sizeof(struct chan_info));
+		ci = chan_info_new();
 		ci->name = strdup(chan);
 		ci->key = key ? strdup(key) : NULL;
+		ci->backlog = 1;
 		hash_insert(&LINK(ircs)->chan_infos, chan, ci);
 		list_add_last(&LINK(ircs)->chan_infos_order, ci);
 	} else {
@@ -1035,8 +1060,8 @@ static int irc_dispatch_trust_client(struct link_client *ic, struct line *line)
 }
 #endif
 
-int irc_cli_bip(struct link_client *ic, struct line *line);
-static int irc_dispatch_client(struct link_client *ic, struct line *line)
+static int irc_dispatch_client(bip_t *bip, struct link_client *ic,
+		struct line *line)
 {
 	int r = OK_COPY;
 	if (line->elemc == 0)
@@ -1053,17 +1078,17 @@ static int irc_dispatch_client(struct link_client *ic, struct line *line)
 				"before sending commands\r\n");
 		r = OK_FORGET;
 	} else if (strcasecmp(line->elemv[0], "BIP") == 0) {
-		r = irc_cli_bip(ic, line);
+		r = irc_cli_bip(bip, ic, line);
 	} else if (strcmp(line->elemv[0], "JOIN") == 0) {
 		r = irc_cli_join(ic, line);
 	} else if (strcmp(line->elemv[0], "PART") == 0) {
 		r = irc_cli_part(ic, line);
 	} else if (strcmp(line->elemv[0], "NICK") == 0) {
-		r = irc_cli_nick(ic, line, NULL);
+		r = irc_cli_nick(bip, ic, line);
 	} else if (strcmp(line->elemv[0], "QUIT") == 0) {
 		r = irc_cli_quit(ic, line);
 	} else if (strcmp(line->elemv[0], "PRIVMSG") == 0) {
-		r = irc_cli_privmsg(ic, line);
+		r = irc_cli_privmsg(bip, ic, line);
 	} else if (strcmp(line->elemv[0], "NOTICE") == 0) {
 		r = irc_cli_notice(ic, line);
 	} else if (strcmp(line->elemv[0], "WHO") == 0) {
@@ -1078,7 +1103,9 @@ static int irc_dispatch_client(struct link_client *ic, struct line *line)
 				LINK(ic)->l_server->nick)
 			write_line(CONN(LINK(ic)->l_server), str);
 		else if (LINK(ic)->l_server->nick)
-			WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG",
+			WRITE_LINE2(CONN(ic), P_IRCMASK,
+					(LINK(ic)->user->bip_use_notice ?
+						"NOTICE" : "PRIVMSG"),
 					LINK(ic)->l_server->nick,
 					":Not connected please try again "
 					"later...\r\n");
@@ -1120,7 +1147,7 @@ static void irc_copy_cli(struct link_client *src, struct link_client *dest,
 
 	/* LINK(src) == LINK(dest) */
 	size_t len = strlen(line->elemv[2]) + 5;
-	char *tmp;
+	char *tmp, *oldelem;
 
 	if (len == 0)
 		return;
@@ -1134,53 +1161,55 @@ static void irc_copy_cli(struct link_client *src, struct link_client *dest,
 	/* tricky: */
 	line->elemv[1] = LINK(src)->l_server->nick;
 
-	free(line->elemv[2]);
+	oldelem = line->elemv[2];
 	line->elemv[2] = tmp;
 	str = irc_line_to_string(line);
 	/* end of trick: */
 	line->elemv[1] = line->origin;
+	line->elemv[2] = oldelem;
 	line->origin = NULL;
 	write_line(CONN(dest), str);
+	free(tmp);
 	free(str);
 	return;
 }
 
-static int irc_dispatch_loging_client(struct link_client *ic, struct line *line,
-		list_t *linkl)
+static int irc_dispatch_loging_client(bip_t *bip, struct link_client *ic,
+		struct line *line)
 {
 	if (line->elemc == 0)
 		return ERR_PROTOCOL;
 
 	if (strcmp(line->elemv[0], "NICK") == 0) {
-		return irc_cli_nick(ic, line, linkl);
+		return irc_cli_nick(bip, ic, line);
 	} else if (strcmp(line->elemv[0], "USER") == 0) {
-		return irc_cli_user(ic, line, linkl);
+		return irc_cli_user(bip, ic, line);
 	} else if (strcmp(line->elemv[0], "PASS") == 0) {
-		return irc_cli_pass(ic, line, linkl);
+		return irc_cli_pass(bip, ic, line);
 	}
 	return OK_FORGET;
 }
 
-int irc_dispatch(struct link_any *l, struct line *line, list_t *linkl)
+int irc_dispatch(bip_t *bip, struct link_any *l, struct line *line)
 {
 	switch (TYPE(l)) {
 	case IRC_TYPE_SERVER:
-		return irc_dispatch_server((struct link_server*)l, line);
+		return irc_dispatch_server(bip, (struct link_server*)l, line);
 		break;
 	case IRC_TYPE_CLIENT:
-		return irc_dispatch_client((struct link_client*)l, line);
+		return irc_dispatch_client(bip, (struct link_client*)l, line);
 		break;
 	case IRC_TYPE_LOGING_CLIENT:
-		return irc_dispatch_loging_client((struct link_client*)l,
-				line, linkl);
+		return irc_dispatch_loging_client(bip, (struct link_client*)l,
+				line);
 		break;
-#ifdef HAVE_LIBSSL		
+#ifdef HAVE_LIBSSL
 	case IRC_TYPE_TRUST_CLIENT:
 		return irc_dispatch_trust_client((struct link_client*)l, line);
 		break;
 #endif
 	default:
-		fatal("gn");
+		fatal("gnéééééé");
 	}
 	return ERR_PROTOCOL; /* never reached */
 }
@@ -1410,11 +1439,6 @@ static void channel_free(struct channel *c)
 	if (c->create_ts)
 		free(c->create_ts);
 
-/*
-	char *l;
-	while ((l = (char *)list_remove_first(&c->bans)))
-		free(l);
-*/
 	hash_iterator_t hi;
 	for (hash_it_init(&c->nicks, &hi); hash_it_item(&hi); hash_it_next(&hi))
 		nick_free(hash_it_item(&hi));
@@ -1460,7 +1484,7 @@ static int irc_part(struct link_server *server, struct line *line)
 	free(s_nick);
 
 	log_part(LINK(server)->log, line->origin, s_chan,
-			line->elemc == 3 ? line->elemv[2]:NULL);
+			line->elemc == 3 ? line->elemv[2] : NULL);
 
 	nick_free(nick);
 	return OK_COPY;
@@ -1541,7 +1565,7 @@ static int irc_mode(struct link_server *server, struct line *line)
 	log_mode(LINK(server)->log, line->origin, line->elemv[1],
 			line->elemv[2], line->elemv + 3, line->elemc - 3);
 
-	/*       
+	/*
 	 * MODE -a+b.. #channel args
 	 *         ^            ^
 	 *       mode         cur_arg
@@ -1698,11 +1722,30 @@ static int irc_kick(struct link_server *server, struct line *line)
 	return OK_COPY;
 }
 
+static void irc_privmsg_check_ctcp(struct link_server *server,
+				   struct line *line)
+{
+	if (line->elemc != 3)
+		return;
+
+	if (!line->origin)
+		return;
+
+	char *nick;
+	nick = nick_from_ircmask(line->origin);
+	if (strcmp(line->elemv[2], "\001VERSION\001") == 0) {
+		WRITE_LINE2(CONN(server), NULL, "NOTICE", nick,
+				"\001VERSION bip-" BIP_VERSION "\001");
+	}
+	free(nick);
+}
+
 static int irc_privmsg(struct link_server *server, struct line *line)
 {
 	if (LINK(server)->s_state == IRCS_CONNECTED)
 		log_privmsg(LINK(server)->log, line->origin, line->elemv[1],
 				line->elemv[2]);
+	irc_privmsg_check_ctcp(server, line);
 	return OK_COPY;
 }
 
@@ -1797,12 +1840,18 @@ static int irc_generic_quit(struct link_server *server, struct line *line)
 static void irc_server_startup(struct link_server *ircs)
 {
 	char *nick;
+	char *username, *realname;
 
 	if (LINK(ircs)->s_password)
 		WRITE_LINE1(CONN(ircs), NULL, "PASS", LINK(ircs)->s_password);
 
-	WRITE_LINE4(CONN(ircs), NULL, "USER", LINK(ircs)->user, "0", "*",
-			LINK(ircs)->real_name);
+	username = LINK(ircs)->username;
+	if (!username)
+		username = LINK(ircs)->user->default_username;
+	realname = LINK(ircs)->realname;
+	if (!realname)
+		realname = LINK(ircs)->user->default_realname;
+	WRITE_LINE4(CONN(ircs), NULL, "USER", username, "0", "*", realname);
 
 	nick = ircs->nick;
 	if (LINK(ircs)->away_nick && LINK(ircs)->l_clientc == 0) {
@@ -1814,6 +1863,9 @@ static void irc_server_startup(struct link_server *ircs)
 			|| nick == NULL) {
 		if (nick)
 			free(nick);
+		if (!LINK(ircs)->connect_nick)
+			LINK(ircs)->connect_nick =
+				strdup(LINK(ircs)->user->default_nick);
 		nick = strdup(LINK(ircs)->connect_nick);
 	}
 
@@ -1824,8 +1876,8 @@ static void irc_server_startup(struct link_server *ircs)
 static void server_next(struct link *l)
 {
 	l->cur_server++;
-	if (l->cur_server >= l->serverc)
-		l->cur_server -= l->serverc;
+	if (l->cur_server >= l->network->serverc)
+		l->cur_server = 0;
 }
 
 static struct link_client *irc_accept_new(connection_t *conn)
@@ -1870,7 +1922,6 @@ void server_cleanup(struct link_server *server)
 		CONN(server) = NULL;
 	}
 	irc_lag_init(server);
-
 }
 
 void irc_client_close(struct link_client *ic)
@@ -1894,6 +1945,22 @@ void irc_client_close(struct link_client *ic)
 	}
 }
 
+static void server_setup_reconnect_timer(struct link *link)
+{
+	int timer = 0;
+
+	if (link->last_connection_attempt &&
+			time(NULL) - link->last_connection_attempt
+				< CONN_INTERVAL) {
+		timer = RECONN_TIMER * (link->s_conn_attempt);
+		if (timer > RECONN_TIMER_MAX)
+			timer = RECONN_TIMER_MAX;
+	}
+	mylog(LOG_ERROR, "%s dead, reconnecting in %d seconds", link->name,
+			timer);
+	link->recon_timer = timer;
+}
+
 static void irc_close(struct link_any *l)
 {
 	if (CONN(l)) {
@@ -1902,25 +1969,16 @@ static void irc_close(struct link_any *l)
 	}
 	if (TYPE(l) == IRC_TYPE_SERVER) {
 		/* TODO: free link_server as a whole */
-		int timer = 0;
 		struct link_server *is = (struct link_server *)l;
 
 		if (LINK(is)->s_state == IRCS_CONNECTED)
 			irc_notify_disconnection(is);
-		else
-			LINK(is)->s_conn_attempt++;
 		irc_server_shutdown(is);
 		log_disconnected(LINK(is)->log);
 
 		server_next(LINK(is));
 		server_cleanup(is);
-		if (LINK(is)->last_connection &&
-				time(NULL) - LINK(is)->last_connection
-					< CONN_INTERVAL)
-			timer = RECONN_TIMER * (LINK(is)->s_conn_attempt);
-		mylog(LOG_ERROR, "%s dead, reconnecting in %d seconds",
-				LINK(l)->name, timer);
-		LINK(is)->recon_timer = timer;
+		server_setup_reconnect_timer(LINK(is));
 
 		LINK(is)->l_server = NULL;
 		irc_server_free((struct link_server *)is);
@@ -1960,10 +2018,21 @@ struct link_server *irc_server_new(struct link *link, connection_t *conn)
 
 void irc_server_free(struct link_server *s)
 {
+	if (CONN(s))
+		connection_free(CONN(s));
 	if (s->nick)
 		free(s->nick);
 	if (s->user_mode)
 		free(s->user_mode);
+
+	hash_iterator_t hi;
+	for (hash_it_init(&s->channels, &hi); hash_it_item(&hi);
+			hash_it_next(&hi)) {
+		struct channel *chan = hash_it_item(&hi);
+		channel_free(chan);
+
+	}
+
 	free(s);
 }
 
@@ -1972,26 +2041,38 @@ connection_t *irc_server_connect(struct link *link)
 	struct link_server *ls;
 	connection_t *conn;
 
+	link->s_conn_attempt++;
+
 	mylog(LOG_INFO, "Connecting user '%s' to network '%s' using server "
-		"%s:%d", link->username, link->name,
-		link->serverv[link->cur_server]->host,
-		link->serverv[link->cur_server]->port);
-	conn = connection_new(link->serverv[link->cur_server]->host,
-				link->serverv[link->cur_server]->port,
+		"%s:%d", link->user->name, link->name,
+		link->network->serverv[link->cur_server].host,
+		link->network->serverv[link->cur_server].port);
+	conn = connection_new(link->network->serverv[link->cur_server].host,
+				link->network->serverv[link->cur_server].port,
 				link->vhost, link->bind_port,
 #ifdef HAVE_LIBSSL
-				link->s_ssl, link->ssl_check_mode,
-				link->ssl_check_store,
+				link->network->ssl, link->ssl_check_mode,
+				link->user->ssl_check_store,
+				link->user->ssl_client_certfile,
 #else
 				0, 0, NULL,
 #endif
 				CONNECT_TIMEOUT);
 	if (!conn)
 		fatal("connection_new");
+	if (conn->handle == -1) {
+		mylog(LOG_INFO, "Cannot connect.");
+		connection_free(conn);
+		return NULL;
+	}
 
 	ls = irc_server_new(link, conn);
-
 	conn->user_data = ls;
+
+	list_add_last(&_bip->conn_list, conn);
+#ifdef HAVE_OIDENTD
+	oidentd_dump(_bip);
+#endif
 	irc_server_startup(ls);
 	return conn;
 }
@@ -2031,59 +2112,44 @@ void irc_server_shutdown(struct link_server *s)
 #define BIP_OIDENTD_END "## END OF AUTOGENERATED STUFF ##\n"
 #define BIP_OIDENTD_END_LENGTH strlen(BIP_OIDENTD_END)
 
-void oidentd_dump(list_t *connl)
+void oidentd_dump(bip_t *bip)
 {
 	list_iterator_t it;
 	FILE *f;
-	char *home, *filename;
 	char *bipstart = NULL, *bipend = NULL;
 	struct stat stats;
 	char tag_written = 0;
 
-	home = getenv("HOME");
-	if (home == NULL) {
-		mylog(LOG_WARN, "Can't get $HOME, not writing oidentd.conf");
-		return;
-	}
-
-	filename = (char *)malloc(strlen(home) + strlen("/.oidentd.conf"));
-	if(filename == NULL)
-		fatal("Out of memory.");
-
-	sprintf(filename, "%s/.oidentd.conf", home);
-
-	if (stat(filename, &stats) == -1) {
+	if (stat(bip->oidentdpath, &stats) == -1) {
 		if (errno == ENOENT) {
-			f = fopen(filename, "w+");
+			f = fopen(bip->oidentdpath, "w+");
 			fchmod(fileno(f), 0644);
 		} else {
-			mylog(LOG_WARN, "Can't open/create %s", filename);
-			free(filename);
+			mylog(LOG_WARN, "Can't open/create %s",
+					bip->oidentdpath);
 			return;
 		}
 	} else {
 		char *content;
-		f = fopen(filename, "r+");
+		f = fopen(bip->oidentdpath, "r+");
 
 		if (!f) {
-			mylog(LOG_WARN, "Can't open/create %s", filename);
-			free(filename);
+			mylog(LOG_WARN, "Can't open/create %s",
+					bip->oidentdpath);
 			return;
 		}
 
 		content = (char *)malloc(stats.st_size + 1);
 
 		if (content == NULL){
-			mylog(LOG_WARN, "oidentd_dump : malloc failed, "
-					"returning");
-			fclose(f);
-			free(filename);
+			fatal("out of memory");
 			return;
 		}
 
 		if (fread(content, 1, stats.st_size, f) !=
 				(size_t)stats.st_size) {
-			mylog(LOG_WARN, "Can't read %s fully", filename);
+			mylog(LOG_WARN, "Can't read %s fully",
+					bip->oidentdpath);
 			free(content);
 			goto clean_oidentd;
 		}
@@ -2098,7 +2164,7 @@ void oidentd_dump(list_t *connl)
 			fseek(f, SEEK_SET, 0);
 			if (ftruncate(fileno(f), 0) == -1) {
 				mylog(LOG_DEBUG, "Can't reset %s size",
-						filename);
+						bip->oidentdpath);
 				free(content);
 				goto clean_oidentd;
 			}
@@ -2116,24 +2182,27 @@ void oidentd_dump(list_t *connl)
 						BIP_OIDENTD_END_LENGTH, f);
 			else
 				mylog(LOG_WARN, "No %s mark found in %s",
-						BIP_OIDENTD_END, filename);
+						BIP_OIDENTD_END,
+						bip->oidentdpath);
 		} else {
 			/* No previous conf */
 			if (stats.st_size != 0 &&
 					content[stats.st_size - 1] != '\n')
 				fprintf(f, "\n");
 		}
+		free(content);
 	}
 
-	for (list_it_init(connl, &it); list_it_item(&it); list_it_next(&it)) {
+	for (list_it_init(&bip->conn_list, &it); list_it_item(&it);
+			list_it_next(&it)) {
 		connection_t *c = list_it_item(&it);
 		struct link_any *la = c->user_data;
-		if (c->connected == CONN_OK && la &&
-				TYPE(la) == IRC_TYPE_SERVER) {
+		if (la && TYPE(la) == IRC_TYPE_SERVER && (
+				c->connected == CONN_OK ||
+				c->connected == CONN_NEED_SSLIZE ||
+				c->connected == CONN_UNTRUSTED)) {
 			struct link_server *ls;
 			struct link *l;
-			char *localip, *remoteip;
-			int localport, remoteport;
 
 			if (!tag_written) {
 				fprintf(f, BIP_OIDENTD_START);
@@ -2143,18 +2212,11 @@ void oidentd_dump(list_t *connl)
 			ls = (struct link_server*)la;
 			l = LINK(ls);
 
-			localip = connection_localip(CONN(ls));
-			localport = connection_localport(CONN(ls));
-			remoteip = connection_remoteip(CONN(ls));
-			remoteport = connection_remoteport(CONN(ls));
-
 			fprintf(f, "to %s fport %d from %s lport %d {\n",
-					remoteip, remoteport, localip,
-					localport);
-			fprintf(f, "\treply \"%s\"\n", l->user);
+					c->remoteip, c->remoteport, c->localip,
+					c->localport);
+			fprintf(f, "\treply \"%s\"\n", l->username);
 			fprintf(f, "}\n");
-			free(localip);
-			free(remoteip);
 		}
 	}
 	if (tag_written)
@@ -2162,13 +2224,9 @@ void oidentd_dump(list_t *connl)
 
 clean_oidentd:
 	fclose(f);
-	free(filename);
 }
 #endif
 
-/* do not use */
-list_t *_connections = NULL;
-
 void timeout_clean_who_counts(list_t *conns)
 {
 	list_iterator_t it;
@@ -2192,251 +2250,197 @@ void timeout_clean_who_counts(list_t *conns)
 	}
 }
 
-struct link_client *reloading_client;
+void bip_init(bip_t *bip)
+{
+	memset(bip, 0, sizeof(bip_t));
+	list_init(&bip->link_list, list_ptr_cmp);
+	list_init(&bip->conn_list, list_ptr_cmp);
+	list_init(&bip->connecting_client_list, list_ptr_cmp);
+
+	hash_init(&bip->users, HASH_NOCASE);
+	hash_init(&bip->networks, HASH_NOCASE);
+}
+
+/* Called each second. */
+void bip_tick(bip_t *bip)
+{
+	static int logflush_timer = 0;
+	struct link *link;
+	list_iterator_t li;
+
+	/* log flushs */
+	if (logflush_timer-- <= 0) {
+		logflush_timer = conf_log_sync_interval;
+		log_flush_all();
+	}
+
+	/* handle tick for links: detect lags or start a reconnection */
+	for (list_it_init(&bip->link_list, &li); (link = list_it_item(&li));
+			list_it_next(&li)) {
+		if (link->l_server) {
+			if (irc_server_lag_compute(link)) {
+				log_ping_timeout(link->log);
+				list_remove(&bip->conn_list,
+						CONN(link->l_server));
+				irc_close((struct link_any *) link->l_server);
+			}
+		} else {
+			if (link->recon_timer == 0) {
+				connection_t *conn;
+				link->last_connection_attempt = time(NULL);
+				conn = irc_server_connect(link);
+				if (!conn)
+					server_setup_reconnect_timer(link);
+			} else {
+				link->recon_timer--;
+			}
+		}
+	}
+
+	/* drop lagging connecting client */
+	for (list_it_init(&bip->connecting_client_list, &li); list_it_item(&li);
+			list_it_next(&li)) {
+		struct link_client *ic = list_it_item(&li);
+		ic->logging_timer++;
+		if (ic->logging_timer > LOGGING_TIMEOUT) {
+			if (CONN(ic))
+				list_remove(&bip->conn_list, CONN(ic));
+			irc_close((struct link_any *)ic);
+			list_it_remove(&li);
+		}
+	}
+
+	/*
+	 * Cleanup lagging or dangling who_count buffers
+	 */
+	timeout_clean_who_counts(&bip->link_list);
+}
+
+void bip_on_event(bip_t *bip, connection_t *conn)
+{
+	struct link_any *lc = (struct link_any *)conn->user_data;
+
+	if (conn == bip->listener) {
+		struct link_client *n = irc_accept_new(conn);
+		if (!n)
+			fatal("Problem while binding local socket");
+		list_add_last(&bip->conn_list, CONN(n));
+		list_add_last(&bip->connecting_client_list, n);
+		return;
+	}
+
+	/* reached only if socket is not listening */
+	int err;
+	list_t *linel = read_lines(conn, &err);
+	if (err) {
+		if (TYPE(lc) == IRC_TYPE_SERVER) {
+			mylog(LOG_ERROR, "read_lines error, closing %s ...",
+					LINK(lc)->name);
+			irc_server_shutdown(LINK(lc)->l_server);
+		} else {
+			mylog(LOG_ERROR, "read_lines error, closing...");
+		}
+		goto prot_err;
+	}
+	if (!linel)
+		return;
+
+	char *line_s;
+	while ((line_s = list_remove_first(linel))) {
+		struct line *line;
+		mylog(LOG_DEBUG, "\"%s\"", line_s);
+		if (*line_s == 0) { /* irssi does that.*/
+			free(line_s);
+			continue;
+		}
+
+		line = irc_line(line_s);
+		if (!line) {
+			mylog(LOG_ERROR, "Error in protocol, closing...");
+			free(line_s);
+			goto prot_err_lines;
+		}
+		int r;
+		r = irc_dispatch(bip, lc, line);
+		irc_line_free(line);
+		free(line_s);
+		if (r == ERR_PROTOCOL) {
+			mylog(LOG_ERROR, "Error in protocol, "
+					"closing...");
+			goto prot_err_lines;
+		}
+		if (r == ERR_AUTH)
+			goto prot_err_lines;
+		/* XXX: not real error */
+		if (r == OK_CLOSE)
+			goto prot_err_lines;
+
+	}
+	list_free(linel);
+	return;
+prot_err_lines:
+	while ((line_s = list_remove_first(linel)))
+		free(line_s);
+prot_err:
+	list_remove(&bip->conn_list, conn);
+	if (linel)
+		list_free(linel);
+	if (lc) {
+		if (TYPE(lc) == IRC_TYPE_LOGING_CLIENT)
+			list_remove(&bip->connecting_client_list, lc);
+		irc_close(lc);
+	}
+}
+
 /*
  * The main loop
  * inc is the incoming connection, clientl list a list of client struct that
  * represent the accepcted credentials
  */
-void irc_main(connection_t *inc, list_t *ll)
+void irc_main(bip_t *bip)
 {
-	list_t reconnectl;
-	list_t timerwaitl;
-	list_t connl;
-	list_t connecting_c;
-	list_t connected_c;
 	int timeleft = 1000;
-	int logflush_timer = conf_log_sync_interval;
 
-	list_init(&reconnectl, NULL);
-	list_init(&timerwaitl, NULL);
-	list_init(&connl, list_ptr_cmp);
-	list_init(&connecting_c, list_ptr_cmp);
-	list_init(&connected_c, list_ptr_cmp);
+	if (bip->reloading_client) {
+		char *l;
 
-	_connections = &connl;
-
-	/* XXX: This one MUST be first */
-	list_add_first(&connl, inc);
+		while ((l = list_remove_first(&bip->errors)))
+			bip_notify(bip->reloading_client, l);
+		bip->reloading_client = NULL;
+	}
 
 	/*
-	 * Merge with already connected data, happens on SIGHUP
+	 * If the list is empty, we are starting. Otherwise we are reloading,
+	 * and conn_list is kept accross reloads.
 	 */
-	list_iterator_t it;
-	for (list_it_init(ll, &it); list_it_item(&it); list_it_next(&it)) {
-		struct link *link = list_it_item(&it);
-		if (link->l_server)
-			list_add_last(&connl, CONN(link->l_server));
-		else
-			list_add_last(&reconnectl, link);
-		if (link->l_clientc) {
-			int i;
-			for (i = 0; i < link->l_clientc; i++) {
-				struct link_client *c;
-				c = link->l_clientv[i];
-				list_add_last(&connl, CONN(c));
-				if (TYPE(c) == IRC_TYPE_LOGING_CLIENT)
-					list_add_last(&connecting_c, c);
-				if (TYPE(c) == IRC_TYPE_CLIENT)
-					list_add_last(&connected_c, c);
-			}
-		}
-	}
-	if (conf_error && reloading_client) {
-		char *nick;
-		if (LINK(reloading_client)->l_server)
-			nick = LINK(reloading_client)->l_server->nick;
-		else
-			nick = LINK(reloading_client)->prev_nick;
-		WRITE_LINE2(CONN(reloading_client), P_IRCMASK, "PRIVMSG", nick,
-				conf_errstr);
-		reloading_client = NULL;
-	}
-
+	if (list_is_empty(&bip->conn_list))
+		list_add_first(&bip->conn_list, bip->listener);
 
 	while (!sighup) {
-		struct link *link;
 		connection_t *conn;
 
 		if (timeleft == 0) {
 			/*
 			 * Compute timeouts for next reconnections and lagouts
 			 */
-			static int throttle_prot = S_CONN_DELAY - 1;
 
 			timeleft = 1000;
-
-			if (++throttle_prot == S_CONN_DELAY) {
-				throttle_prot = 0;
-
-				/* Lauch one reconnection at a time */
-				if ((link = list_remove_first(&reconnectl))) {
-					conn = irc_server_connect(link);
-					list_add_last(&connl, conn);
-					link->last_connection = time(NULL);
-				}
-			}
-
-			/* log flushs */
-			if (logflush_timer-- <= 0) {
-				logflush_timer = conf_log_sync_interval;
-				log_flush_all();
-			}
-
-			/* reconnects */
-			list_iterator_t li;
-			for (list_it_init(&timerwaitl, &li); list_it_item(&li);
-					list_it_next(&li)) {
-				struct link *l = list_it_item(&li);
-				if (l->recon_timer <= 0) {
-					list_it_remove(&li);
-					list_add_last(&reconnectl, l);
-				} else {
-					l->recon_timer--;
-				}
-			}
-
-			/* server lags */
-			for (list_it_init(ll, &li); list_it_item(&li);
-					list_it_next(&li)) {
-				struct link *l = list_it_item(&li);
-				if(l->l_server && irc_server_lag_compute(l)) {
-					log_ping_timeout(l->log);
-					list_remove(&connl, CONN(l->l_server));
-					irc_close((struct link_any *)
-							l->l_server);
-					list_add_last(&timerwaitl, l);
-				}
-			}
-
-			/* drop lagging connecting client */
-			for (list_it_init(&connecting_c, &li);
-					list_it_item(&li); list_it_next(&li)) {
-				struct link_client *ic = list_it_item(&li);
-				ic->logging_timer++;
-				if (ic->logging_timer > LOGGING_TIMEOUT) {
-					list_remove(&connl, CONN(ic));
-					irc_close((struct link_any *)ic);
-					list_it_remove(&li);
-				}
-			}
-
-			/*
-			 * Cleanup lagging or dangling who_count buffers
-			 */
-			timeout_clean_who_counts(ll);
+			bip_tick(bip);
 		}
 
 		int nc;
 		/* Da main loop */
-		list_t *ready = wait_event(&connl, &timeleft, &nc);
+		list_t *ready = wait_event(&bip->conn_list, &timeleft, &nc);
 #ifdef HAVE_OIDENTD
 		if (nc)
-			oidentd_dump(&connl);
+			oidentd_dump(bip);
 #endif
-		while ((conn = list_remove_first(ready))) {
-			struct link_any *lc =
-				(struct link_any *)conn->user_data;
-
-			if (conn == inc) {
-				struct link_client *n = irc_accept_new(conn);
-				if (!n)
-					fatal("Problem while binding local"
-							" socket");
-				list_add_last(&connl, CONN(n));
-				list_add_last(&connecting_c, n);
-				continue;
-			}
-
-			/* reached only if socket is not listening */
-			int err;
-			list_t *linel = read_lines(conn, &err);
-			if (err) {
-				if (TYPE(lc) == IRC_TYPE_SERVER) {
-					mylog(LOG_ERROR, "read_lines error, "
-						"closing %s ...",
-						LINK(lc)->name);
-					irc_server_shutdown(LINK(lc)->l_server);
-				} else {
-					mylog(LOG_ERROR, "read_lines error, "
-						"closing...");
-				}
-				goto prot_err;
-			}
-			if (!linel)
-				continue;
-
-			char *line_s;
-			while ((line_s = list_remove_first(linel))) {
-				struct line *line;
-				mylog(LOG_DEBUG, "\"%s\"", line_s);
-				if (*line_s == 0) { /* irssi does that.*/
-					free(line_s);
-					continue;
-				}
-
-				line = irc_line(line_s);
-				if (!line) {
-					mylog(LOG_ERROR, "Error in protocol, "
-							"closing...");
-					free(line_s);
-					goto prot_err_lines;
-				}
-				int r;
-				int oldtype = TYPE(lc);
-				r = irc_dispatch((struct link_any*)
-						conn->user_data, line, ll);
-				irc_line_free(line);
-				free(line_s);
-				if (r == ERR_PROTOCOL) {
-					mylog(LOG_ERROR, "Error in protocol, "
-							"closing...");
-					goto prot_err_lines;
-				}
-				if (r == ERR_AUTH)
-					goto prot_err_lines;
-				/* XXX: not real error */
-				if (r == OK_CLOSE)
-					goto prot_err_lines;
-
-				if (oldtype == IRC_TYPE_LOGING_CLIENT &&
-						TYPE(lc) == IRC_TYPE_CLIENT) {
-					list_remove(&connecting_c, lc);
-					list_add_last(&connected_c, lc);
-				}
-			}
-			list_free(linel);
-			continue;
-prot_err_lines:
-			while ((line_s = list_remove_first(linel)))
-				free(line_s);
-prot_err:
-			list_remove(&connl, conn);
-			if (linel)
-				list_free(linel);
-			if (lc) {
-				if (TYPE(lc) == IRC_TYPE_CLIENT)
-					list_remove(&connected_c, lc);
-				if (TYPE(lc) == IRC_TYPE_LOGING_CLIENT)
-					list_remove(&connecting_c, lc);
-				if (TYPE(lc) == IRC_TYPE_SERVER)
-					list_add_last(&timerwaitl, LINK(lc));
-				irc_close(lc);
-			}
-		}
+		while ((conn = list_remove_first(ready)))
+			bip_on_event(bip, conn);
 		list_free(ready);
 	}
-	while (list_remove_first(&connecting_c))
+	while (list_remove_first(&bip->connecting_client_list))
 		;
-	while (list_remove_first(&connected_c))
-		;
-	while (list_remove_first(&connl))
-		;
-	while (list_remove_first(&timerwaitl))
-		;
-	while (list_remove_first(&reconnectl))
-		;
-	_connections = NULL;
 	return;
 }
 
@@ -2451,12 +2455,6 @@ void irc_client_free(struct link_client *cli)
 	free(cli);
 }
 
-/*
-void irc_server_free(struct link_server *is)
-{
-	free(is);
-}
-*/
 struct link *irc_link_new()
 {
 	struct link *link;
@@ -2464,9 +2462,57 @@ struct link *irc_link_new()
 	if (!link)
 		fatal("calloc");
 
+	link->l_server = NULL;
 	hash_init(&link->chan_infos, HASH_NOCASE);
 	list_init(&link->chan_infos_order, list_ptr_cmp);
 	list_init(&link->on_connect_send, list_ptr_cmp);
 	return link;
 }
 
+void link_kill(bip_t *bip, struct link *link)
+{
+	/* in case in never got connected */
+	if (link->l_server) {
+		list_remove(&bip->conn_list, CONN(link->l_server));
+		server_cleanup(link->l_server);
+		irc_server_free(link->l_server);
+	}
+	while (link->l_clientc) {
+		struct link_client *lc = link->l_clientv[0];
+		if (lc == bip->reloading_client)
+			bip->reloading_client = NULL;
+		list_remove(&bip->conn_list, CONN(lc));
+		unbind_from_link(lc);
+		irc_client_free(lc);
+	}
+
+	hash_remove(&link->user->connections, link->name);
+	free(link->name);
+	log_free(link->log);
+	MAYFREE(link->prev_nick);
+	MAYFREE(link->cli_nick);
+
+	void *p;
+	while ((p = list_remove_first(&link->init_strings)))
+		free(p);
+	while ((p = list_remove_first(&link->on_connect_send)))
+		free(p);
+	MAYFREE(link->no_client_away_msg);
+	MAYFREE(link->away_nick);
+	hash_clean(&link->chan_infos);
+
+	struct chan_infos *ci;
+	while ((ci = list_remove_first(&link->chan_infos_order)))
+		free(ci);
+
+	MAYFREE(link->username);
+	MAYFREE(link->realname);
+	MAYFREE(link->s_password);
+	MAYFREE(link->connect_nick);
+	MAYFREE(link->vhost);
+#ifdef HAVE_LIBSSL
+	sk_X509_free(link->untrusted_certs);
+#endif
+	free(link);
+}
+
